#include "string_representation.h"
#include "inner_storage.h"
#include <assert.h>
#include "utf8.h"

namespace json {
    void escape_string(std::string& dest, const std::string& src) {
        if (!isUtf8String(src)) {
            throw misuse("Can't stringify json string value that is not a utf-8 string");
        }
        dest += '"';
        for (char ch: src) {
            if (ch == '"' || ch == '\\' || (0 <= ch && ch <= 0x1f)) {
                char buf[10];
                snprintf(buf, 10, "\\u%04x", ch);
                dest += buf;
            } else {
                dest += ch;
            }
        }
        dest += '"';
    }

    int try_append_rep_childless(const JSON& obj, uint32_t mode, std::string& ret_ans) {
#define azaza(sym) case sym ## _symbol: ret_ans += #sym; return 0;
        switch (obj.type) {
            azaza(null)
            azaza(false)
            azaza(true)
            case integer:
                ret_ans += obj.asInteger().to_string();
                return 0;
            case string:
                escape_string(ret_ans, obj.asString());
                return 0;
            case array:
                if (obj.asArray().empty()) {
                    ret_ans += "[]";
                    return 0;
                }
                return -1;
            case dictionary:
                if (obj.asDictionary().empty()) {
                    ret_ans += "{}";
                    return 0;
                }
                return -1;
            default:
                assert(false);
        }
    }

    void indent(int depth, std::string& res) {
        for (int i = 0; i < depth; i++)
            res += "    ";
    }

    /* Returns node.it == node.data.end(). If true is returned, cur is switched to parent */
    template<typename ObjData>
    bool print_iter_pref(ObjData& node, std::string& result, int& cur_depth, const JSON*& cur,
        char container_begin, char container_end, uint32_t mode) {
        if (node.it == node.data.begin()) {
            cur_depth++;
            result += container_begin;
            if (mode == print_pretty) {
                result += "\n";
                indent(cur_depth, result);
            }
        } else if (node.it == node.data.end()) {
            cur_depth--;
            if (mode == print_pretty) {
                result += '\n';
                indent(cur_depth, result);
            }
            result += container_end;
            cur = (const JSON*)node.wayback;
            return true;
        } else {
            result += ',';
            if (mode == print_pretty) {
                result += '\n';
                indent(cur_depth, result);
            }
        }
        return false;
    }

    std::string generate_str(const JSON &obj, uint32_t mode) {
        int ret;
        std::string result;
        ret = try_append_rep_childless(obj, mode, result);
        if (ret < 0) {
            int cur_depth = 0;
            const JSON* cur = &obj;
            setup_nr_iterators(obj, NULL);
            while (cur) {
                if (cur->isArray()) {
                    ArrayData& node = *static_cast<ArrayData*>(cur->value);
                    proc_child_1:
                    if (print_iter_pref(node, result, cur_depth, cur, '[', ']', mode))
                        continue;
                    JSON& child = *node.it;
                    ++node.it;
                    ret = try_append_rep_childless(child, mode, result);
                    if (ret == 0) {
                        goto proc_child_1;
                    }
                    setup_nr_iterators(child, (void*)cur);
                    cur = &child;
                } else if (cur->isDictionary()) {
                    DictionaryData& node = *static_cast<DictionaryData*>(cur->value);
                    proc_child_2:
                    if (print_iter_pref(node, result, cur_depth, cur, '{', '}', mode))
                        continue;
                    escape_string(result, node.it->first);
                    result += ':';
                    if (mode == print_pretty)
                        result += ' ';
                    JSON& child = node.it->second;
                    ++node.it;
                    ret = try_append_rep_childless(child, mode, result);
                    if (ret == 0)
                        goto proc_child_2;
                    setup_nr_iterators(child, (void*)cur);
                    cur = &child;
                }
            }
        }
        if (mode == print_pretty) {
            result += "\n";
        }
        return result;
    }
}
